;;; Dezyne --- Dezyne command line tools
;;;
;;; Copyright © 2020 Jan Nieuwenhuizen <janneke@gnu.org>
;;;
;;; This file is part of Dezyne.
;;;
;;; Dezyne is free software: you can redistribute it and/or modify it
;;; under the terms of the GNU Affero General Public License as
;;; published by the Free Software Foundation, either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; Dezyne is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Affero General Public License for more details.
;;;
;;; You should have received a copy of the GNU Affero General Public
;;; License along with Dezyne.  If not, see <http://www.gnu.org/licenses/>.
;;;
;;; Commentary:
;;;
;;; Tests for the makreel module.
;;;
;;; Code:

(define-module (test gaiag makreel)
  #:use-module (gaiag parse peg)
  #:use-module (gaiag parse ast)
  #:use-module (gaiag makreel)
  #:use-module (oop goops)
  #:use-module (gaiag goops)
  #:use-module (srfi srfi-64)
  #:use-module (test gaiag automake))

(define (string->ast string)
  ;; XXX TODO: provide sensible interface to parse-string that does not
  ;; require ->ast, string duplication or exit (1).
  (false-if-exception (parse-root->ast (peg:parse string "<string>") #:string string)))

(test-begin "makreel")

(test-assert "makreel dummy"
  #t)

(test-assert "makreel parse"
  (string->ast "interface i {in void e();behaviour {on e: {}}}"))

(test-assert "makreel tail-call"
  (let* ((ast (string->ast "interface i {in void e();behaviour {void f () {f ();} on e: {}}}"))
         (ast (makreel:om ast))
         (call (car (tree-collect (is? <call>) ast))))
    (.last? call)))

(test-assert "makreel tail-call other"
  (let* ((ast (string->ast "interface i {in void e();behaviour { void g () {} void f () {g ();} on e: {}}}"))
         (ast (makreel:om ast))
         (call (car (tree-collect (is? <call>) ast))))
    (.last? call)))

(test-end)
